==> atmsupport-1.0.tar.gz: atmsupport-1.0/libats/ats_func.c <==
/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/libats/ats_func.c,v 1.7 2003/09/25 13:47:59 hbb Exp $
 */

#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <err.h>

#include "ats.h"

/* list of known functions */
static SLIST_HEAD(, func) funcs = SLIST_HEAD_INITIALIZER(funcs);

/*
 * Dump all functions
 */
void
funcs_dump(void)
{
	struct func *f;
	struct string *str;

	SLIST_FOREACH(f, &funcs, link) {
		printf("# %s, %u\n", f->def.name, f->def.lno);
		printf("<< %s", f->name);
		TAILQ_FOREACH(str, &f->args, link)
			printf(" %s", str->string);
		printf("\n");
		TAILQ_FOREACH(str, &f->lines, link)
			printf("%s\n", str->string);
		printf(">>\n");
	}
}

/*
 * Define a function.
 */
struct func *
func_define(const char *name, const struct ipos *def,
    struct string_list *lines, struct string_list *args)
{
	struct func *f;
	struct string *s;
	const char *p, *t;

	SLIST_FOREACH(f, &funcs, link)
		if (strcmp(f->name, name) == 0)
			break;

	if (f != NULL)
		err_file("redefinition of function '%s'", name);

	if ((f = malloc(sizeof(*f))) == NULL)
		err(1, NULL);
	memset(f, 0, sizeof(*f));

	f->name = name;
	if ((f->def.name = strdup(def->name)) == NULL)
		err(1, NULL);
	f->def.lno = def->lno;

	TAILQ_INIT(&f->lines);
	TAILQ_CONCAT(&f->lines, lines, link);

	TAILQ_INIT(&f->args);
	TAILQ_CONCAT(&f->args, args, link);

	TAILQ_FOREACH(s, args, link) {
		f->nargs++;
		p = s->string;
		if (*p == '&')
			p++;
		t = p;
		while (*p != '\0' && isascii(*p) && isalpha(*p))
			p++;
		if (*p != '\0' || p == t)
			err_file("bad formal argument '%s'", s->string);
	}

	SLIST_INSERT_HEAD(&funcs, f, link);

	return (f);
}

/*
 * Define the arguments for a function
 */
static void
define_arguments(const struct func *f, const struct string_list *act_list)
{
	struct var *ref;
	const struct string *form, *act;

	var_level++;

	form = TAILQ_FIRST(&f->args);
	TAILQ_FOREACH(act, act_list, link) {
		if (form == NULL)
			err_file("%s -- too many actual args", f->name);
		if (form->string[0] == '&') {
			ref = var_find(act->string, NULL);
			if (ref == NULL)
				err_file("%s: referenced variable not found",
				    act->string);
			var_define(form->string + 1, NULL, 1, ref);
		} else
			var_define(form->string, act->string, 0, NULL);
		form = TAILQ_NEXT(form, link);
	}
	if (form != NULL)
		err_file("%s -- not enough args", f->name);
}

/*
 * Call the named function
 */
void
func_call(const char *name, const struct string_list *act_list)
{
	struct func *f;

	SLIST_FOREACH(f, &funcs, link)
		if (strcmp(f->name, name) == 0)
			break;
	if (f == NULL)
		err_file("unknown function '%s'", name);
	if (f->repl)
		err_file("cannot call a string replacement '%s'", name);

	if (ISVERB('f'))
		printf("calling '%s'\n", name);

	define_arguments(f, act_list);

	input_open_func(f);
}

/*
 * Define a replacement
 */
void
replace_define(const char *name, const struct ipos *def,
    struct string_list *args, const char *repl)
{
	struct func *f;
	struct string *s;
	const char *p, *t;

	SLIST_FOREACH(f, &funcs, link)
		if (strcmp(f->name, name) == 0)
			break;

	if (f != NULL)
		err_file("redefinition of function '%s'", name);

	if ((f = malloc(sizeof(*f))) == NULL)
		err(1, NULL);
	memset(f, 0, sizeof(*f));

	f->name = name;
	if ((f->def.name = strdup(def->name)) == NULL)
		err(1, NULL);
	f->def.lno = def->lno;
	f->repl = 1;

	TAILQ_INIT(&f->args);
	TAILQ_CONCAT(&f->args, args, link);

	if ((f->str = strdup(repl)) == NULL)
		err(1, NULL);

	TAILQ_FOREACH(s, args, link) {
		f->nargs++;
		p = s->string;
		if (*p == '&')
			p++;
		t = p;
		while (*p != '\0' && isascii(*p) && isalpha(*p))
			p++;
		if (*p != '\0' || p == t)
			err_file("bad formal argument '%s'", s->string);
	}

	SLIST_INSERT_HEAD(&funcs, f, link);
}

/*
 * The current pointer of the line points to the '$' starting a string
 * substitution. Do this substitution and leave the pointer at the end
 * of the substituted string.
 */
struct line *
str_subst(const char *str, const char **end)
{
	const char *ptr, *p;
	const char *arg;
	const struct func *f;
	struct string_list args;
	struct string *a;
	struct line *n;
	u_int level;
	int out;
	int c;

	/*
 	 * collect the function name
	 */
	ptr = str;
	for (;;) {
		if ((c = *ptr) == '\0')
			err_file("eol in function name");
		if (!isascii(c))
			err_file("bad char in variable name %#02x", c);
		if (!isalnum(c) && c != '_')
			break;
		ptr++;
	}
	if (ptr == str)
		err_file("function name expected");

	SLIST_FOREACH(f, &funcs, link)
		if (strlen(f->name) == (size_t)(ptr - str) &&
		    strncmp(f->name, str, ptr - str) == 0)
			break;
	if (f == NULL)
		err_file("unknown function '%.*s", (int)(ptr - str), str);
	if (!f->repl)
		err_file("'%.*s' is not a string function",
		    (int)(ptr - str), str);

	for (;;) {
		if ((c = *ptr++) == '\0')
			err_file("expecting '('");
		if (c == '(')
			break;
		if (!isascii(c))
			err_file("bad char %#02x", c);
		if (!isblank(c))
			err_file("expecting '('");
	}

	/* collect to matching ')' */
	level = 1;
	for (p = ptr;; p++) {
		if (*p == '\0')
			err_file("eol in expression");
		if (*p == '(')
			level++;
		else if (*p == ')' && --level == 0) {
			p++;
			break;
		}
	}

	n = line_create(p - ptr + 1);
	strncpy(n->line, ptr, p - ptr);
	n->line[p - ptr] = '\0';

	/* run substitution on this string */
	substitute_line(n);

	/*
	 * Collect actual arguments
	 */
	TAILQ_INIT(&args);

	if (*n->ptr == ')') {
		n->ptr++;

	} else {
		out = 0;
		do {
			arg = n->ptr;
			level = 0;
			for (;;) {
				if (*n->ptr == '\0')
					err_file("eol in arg list");
				else if (*n->ptr == ',') {
					if (level == 0)
						break;
				} else if (*n->ptr == ')') {
					if (level == 0) {
						out = 1;
						break;
					}
					level--;
				} else if (*n->ptr == '(')
					level++;
				n->ptr++;
			}
			a = string_saven(arg, input->pos.lno, n->ptr - arg);
			TAILQ_INSERT_TAIL(&args, a, link);
			n->ptr++;
		} while (!out);
	}
	if (*n->ptr != '\0')
		err_file("ups - mismatch in function call");

	define_arguments(f, &args);
	line_destroy(n);

	/*
	 * construct a local copy of the replacement string
	 */
	n = line_create(strlen(f->str) + 1);
	strcpy(n->line, f->str);

	/* make substitutions on this copy */
	substitute_line(n);
	var_pop(var_level--);

	if (ISVERB('s'))
		printf("string substitution: '%.*s' -> '%s'\n",
		    (int)(p - str + 1), str - 1, n->line);
	*end = p;
	return (n);
}

==> atmsupport-1.4.tar.gz: atmsupport-1.4/libats/ats_func.c <==
/*
 * Copyright (c) 2003-2003
 *	Fraunhofer Institute for Open Communication Systems (FhG Fokus).
 * 	All rights reserved.
 *
 * Author: Hartmut Brandt <harti@freebsd.org>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Begemot: atmsupport/libats/ats_func.c,v 1.8 2004/07/08 13:33:35 brandt Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "ats.h"

/* list of known functions */
static LIST_HEAD(, func) funcs = LIST_HEAD_INITIALIZER(funcs);

/*
 * Dump all functions
 */
void
funcs_dump(void)
{
	struct func *f;
	struct string *str;

	LIST_FOREACH(f, &funcs, link) {
		printf("# %s, %u\n", f->def.name, f->def.lno);
		printf("<< %s", f->name);
		TAILQ_FOREACH(str, &f->args, link)
			printf(" %s", str->string);
		printf("\n");
		TAILQ_FOREACH(str, &f->lines, link)
			printf("%s\n", str->string);
		printf(">>\n");
	}
}

/*
 * Define a function.
 */
struct func *
func_define(const char *name, const struct ipos *def,
    struct string_list *lines, struct string_list *args, int stat)
{
	struct func *f;
	struct string *s;
	const char *p, *t;

	LIST_FOREACH(f, &funcs, link)
		if (strcmp(f->name, name) == 0)
			break;

	if (f != NULL)
		err_file("redefinition of function '%s'", name);

	f = ats_alloc(sizeof(*f));

	f->name = name;
	f->def.name = ats_strdup(def->name);
	f->def.lno = def->lno;
	if ((f->stat = stat) != 0)
		f->file_level = input_file_level;
	else
		f->file_level = 0;

	TAILQ_INIT(&f->lines);
	TAILQ_CONCAT(&f->lines, lines, link);

	TAILQ_INIT(&f->args);
	TAILQ_CONCAT(&f->args, args, link);

	TAILQ_FOREACH(s, args, link) {
		f->nargs++;
		p = s->string;
		if (*p == '&')
			p++;
		t = p;
		while (*p != '\0' && isascii(*p) && isalpha(*p))
			p++;
		if (*p != '\0' || p == t)
			err_file("bad formal argument '%s'", s->string);
	}

	LIST_INSERT_HEAD(&funcs, f, link);

	return (f);
}

/*
 * Define the arguments for a function
 */
static void
define_arguments(const struct func *f, const struct string_list *act_list)
{
	struct var *ref;
	const struct string *form, *act;

	var_level++;

	form = TAILQ_FIRST(&f->args);
	TAILQ_FOREACH(act, act_list, link) {
		if (form == NULL)
			err_file("%s -- too many actual args", f->name);
		if (form->string[0] == '&') {
			ref = var_find(act->string, NULL);
			if (ref == NULL)
				err_file("%s: referenced variable not found",
				    act->string);
			var_define(form->string + 1, NULL, 1, ref);
		} else
			var_define(form->string, act->string, 0, NULL);
		form = TAILQ_NEXT(form, link);
	}
	if (form != NULL)
		err_file("%s -- not enough args", f->name);
}

/*
 * Call the named function
 */
void
func_call(const char *name, const struct string_list *act_list)
{
	struct func *f;

	LIST_FOREACH(f, &funcs, link)
		if (strcmp(f->name, name) == 0)
			break;
	if (f == NULL)
		err_file("unknown function '%s'", name);
	if (f->repl)
		err_file("cannot call a string replacement '%s'", name);

	if (ISVERB('f'))
		printf("calling '%s'\n", name);

	define_arguments(f, act_list);

	input_open_func(f);
}

/*
 * Define a replacement
 */
void
replace_define(const char *name, const struct ipos *def,
    struct string_list *args, const char *repl, int stat)
{
	struct func *f;
	struct string *s;
	const char *p, *t;

	LIST_FOREACH(f, &funcs, link)
		if (strcmp(f->name, name) == 0)
			break;

	if (f != NULL)
		err_file("redefinition of function '%s'", name);

	f = malloc(sizeof(*f));

	f->name = name;
	f->def.name = ats_strdup(def->name);
	f->def.lno = def->lno;
	f->repl = 1;
	if ((f->stat = stat) != 0)
		f->file_level = input_file_level;
	else
		f->file_level = 0;

	TAILQ_INIT(&f->args);
	TAILQ_CONCAT(&f->args, args, link);

	f->str = ats_strdup(repl);

	TAILQ_FOREACH(s, args, link) {
		f->nargs++;
		p = s->string;
		if (*p == '&')
			p++;
		t = p;
		while (*p != '\0' && isascii(*p) && isalpha(*p))
			p++;
		if (*p != '\0' || p == t)
			err_file("bad formal argument '%s'", s->string);
	}

	LIST_INSERT_HEAD(&funcs, f, link);
}

/*
 * The current pointer of the line points to the '$' starting a string
 * substitution. Do this substitution and leave the pointer at the end
 * of the substituted string.
 */
struct line *
str_subst(const char *str, const char **end)
{
	const char *ptr, *p;
	const char *arg;
	const struct func *f;
	struct string_list args;
	struct string *a;
	struct line *n;
	u_int level;
	int out;
	int c;

	/*
 	 * collect the function name
	 */
	ptr = str;
	for (;;) {
		if ((c = *ptr) == '\0')
			err_file("eol in function name");
		if (!isascii(c))
			err_file("bad char in variable name %#02x", c);
		if (!isalnum(c) && c != '_')
			break;
		ptr++;
	}
	if (ptr == str)
		err_file("function name expected");

	LIST_FOREACH(f, &funcs, link)
		if (strlen(f->name) == (size_t)(ptr - str) &&
		    strncmp(f->name, str, ptr - str) == 0)
			break;
	if (f == NULL)
		err_file("unknown function '%.*s", (int)(ptr - str), str);
	if (!f->repl)
		err_file("'%.*s' is not a string function",
		    (int)(ptr - str), str);

	for (;;) {
		if ((c = *ptr++) == '\0')
			err_file("expecting '('");
		if (c == '(')
			break;
		if (!isascii(c))
			err_file("bad char %#02x", c);
		if (!isblank(c))
			err_file("expecting '('");
	}

	/* collect to matching ')' */
	level = 1;
	for (p = ptr;; p++) {
		if (*p == '\0')
			err_file("eol in expression");
		if (*p == '(')
			level++;
		else if (*p == ')' && --level == 0) {
			p++;
			break;
		}
	}

	n = line_create(p - ptr + 1);
	strncpy(n->line, ptr, p - ptr);
	n->line[p - ptr] = '\0';

	/* run substitution on this string */
	substitute_line(n);

	/*
	 * Collect actual arguments
	 */
	TAILQ_INIT(&args);

	if (*n->ptr == ')') {
		n->ptr++;

	} else {
		out = 0;
		do {
			arg = n->ptr;
			level = 0;
			for (;;) {
				if (*n->ptr == '\0')
					err_file("eol in arg list");
				else if (*n->ptr == ',') {
					if (level == 0)
						break;
				} else if (*n->ptr == ')') {
					if (level == 0) {
						out = 1;
						break;
					}
					level--;
				} else if (*n->ptr == '(')
					level++;
				n->ptr++;
			}
			a = string_saven(arg, input->pos.lno, n->ptr - arg);
			TAILQ_INSERT_TAIL(&args, a, link);
			n->ptr++;
		} while (!out);
	}
	if (*n->ptr != '\0')
		err_file("ups - mismatch in function call");

	define_arguments(f, &args);
	line_destroy(n);

	/*
	 * construct a local copy of the replacement string
	 */
	n = line_create(strlen(f->str) + 1);
	strcpy(n->line, f->str);

	/* make substitutions on this copy */
	substitute_line(n);
	var_pop(var_level--);

	if (ISVERB('s'))
		printf("string substitution: '%.*s' -> '%s'\n",
		    (int)(p - str + 1), str - 1, n->line);
	*end = p;
	return (n);
}

void
func_destroy(struct func *f)
{

	LIST_REMOVE(f, link);
	free((void *)(uintptr_t)f->name);	/* argh.... */
	free((void *)(uintptr_t)f->str);
	stringlist_free(&f->lines);
	stringlist_free(&f->args);
	free(f);
}

void
func_flush_static(void)
{
	struct func *f, *t;

	LIST_FOREACH_SAFE(f, &funcs, link, t)
		if (f->stat && f->file_level >= input_file_level)
			func_destroy(f);
}
